@@ -13,7 +13,7 @@
  * You should have received a copy of the GNU General Public License
  * along with the hkl library.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Copyright (C) 2003-2021 Synchrotron SOLEIL
+ * Copyright (C) 2003-2022 Synchrotron SOLEIL
  *                         L'Orme des Merisiers Saint-Aubin
  *                         BP 48 91192 GIF-sur-YVETTE CEDEX
  *
@@ -96,7 +96,7 @@ struct imxpad_t {
 #define IMXPAD(pixel_size_, chip_w_, chip_h_) (struct imxpad_t) \
         {.square=SQUARE(pixel_size_), .chip_w=chip_w_, .chip_h=chip_h_}
 
-struct dectris_t {
+struct tiling_t {
         struct square_t square;
         int module_width;
         int module_height;
@@ -104,7 +104,7 @@ struct dectris_t {
         int gap_height;
 };
 
-#define DECTRIS(module_width_, module_height_, gap_width_, gap_height_, pixel_size_) (struct dectris_t) \
+#define TILING(module_width_, module_height_, gap_width_, gap_height_, pixel_size_) (struct tiling_t) \
         {.square=SQUARE(pixel_size_), .module_width=module_width_, .module_height=module_height_, .gap_width=gap_width_, .gap_height=gap_height_}
 
 datatype(
@@ -112,9 +112,10 @@ datatype(
         (ImXpadS70, struct imxpad_t),
         (ImXpadS140, struct imxpad_t),
         (XpadFlatCorrected, struct square_t),
-        (Eiger1M, struct dectris_t),
+        (Eiger1M, struct tiling_t),
         (Ufxc, struct square_t),
-        (Merlin, struct square_t)
+        (Merlin, struct square_t),
+        (MerlinMedipix3RXQuad, struct tiling_t)
         );
 
 struct detector_t {
@@ -136,11 +137,13 @@ static inline struct detector_t get_detector(HklBinocularsDetectorEnum n)
                 DETECTOR(ImXpadS70,
                          SHAPE(560, 120), IMXPAD(130e-6, 80, 120)),
                 DETECTOR(Eiger1M,
-                         SHAPE(1030, 1065), DECTRIS(1030, 514, 10, 37, 75e-6)),
+                         SHAPE(1030, 1065), TILING(1030, 514, 10, 37, 75e-6)),
                 DETECTOR(Ufxc,
                          SHAPE(257, 256), SQUARE(75e-6)),
                 DETECTOR(Merlin,
                          SHAPE(256, 256), SQUARE(55e-6)),
+                DETECTOR(MerlinMedipix3RXQuad,
+                         SHAPE(515, 515), TILING(256, 256, 3, 3, 55e-6)),
         };
         return detectors[n];
 }
@@ -311,28 +314,28 @@ static inline uint8_t *mask_get_xpad_flat_corrected(const struct shape_t *shape)
         return arr;
 }
 
-static inline uint8_t *mask_get_dectris(const struct shape_t *shape,
-                                        const struct dectris_t *dectris)
+static inline uint8_t *mask_get_tiling(const struct shape_t *shape,
+                                       const struct tiling_t *tiling)
 {
         int i;
         uint8_t *arr = no_mask(shape);
 
         /* columns */
-        for(i=dectris->module_width;
+        for(i=tiling->module_width;
             i<shape->width;
-            i=i+dectris->module_width + dectris->gap_width){
+            i=i+tiling->module_width + tiling->gap_width){
                 uint8_t *col = get_col(arr, i);
                 fill_column(col, *shape, 1);
-                replicate_column(col, *shape, dectris->gap_width);
+                replicate_column(col, *shape, tiling->gap_width);
         }
 
         /* rows */
-        for(i=dectris->module_height;
+        for(i=tiling->module_height;
             i<shape->height;
-            i=i+dectris->module_height + dectris->gap_height){
+            i=i+tiling->module_height + tiling->gap_height){
                 uint8_t *row = get_row(arr, *shape, i);
                 fill_row(row, *shape, 1);
-                replicate_row(row, *shape, dectris->gap_height);
+                replicate_row(row, *shape, tiling->gap_height);
         }
 
         return arr;
@@ -387,10 +390,10 @@ static inline void normalize_coordinates(double *arr, const struct shape_t shape
         double *y = y_coordinates(arr, shape);
         double *z = z_coordinates(arr, shape);
 
-        for(int i=0; i>shape_size(shape); ++i)
+        for(int i=0; i<shape_size(shape); ++i)
         {
                 double n = sqrt(x[i] * x[i] + y[i] * y[i] + z[i] * z[i]);
-                if (n > -DBL_MAX){
+                if (n != 0.0){
                         x[i] = x[i] / n;
                         y[i] = y[i] / n;
                         z[i] = z[i] / n;
@@ -445,33 +448,39 @@ void hkl_binoculars_detector_2d_shape_get(HklBinocularsDetectorEnum n,
 double *hkl_binoculars_detector_2d_coordinates_get(HklBinocularsDetectorEnum n)
 {
         const struct detector_t detector = get_detector(n);
+        double *arr = NULL;
 
         match(detector.type){
                 of(ImXpadS70, imxpad){
-                        return coordinates_get_imxpad(&detector.shape,
-                                                      imxpad);
+                        arr = coordinates_get_imxpad(&detector.shape,
+                                                     imxpad);
                 }
                 of(ImXpadS140, imxpad){
-                        return coordinates_get_imxpad(&detector.shape,
-                                                      imxpad);
+                        arr = coordinates_get_imxpad(&detector.shape,
+                                                     imxpad);
                 }
                 of(XpadFlatCorrected, square){
-                        return coordinates_get_square(&detector.shape,
-                                                      square);
+                        arr = coordinates_get_square(&detector.shape,
+                                                     square);
                 }
-                of(Eiger1M, dectris){
-                        return coordinates_get_square(&detector.shape,
-                                                      &dectris->square);
+                of(Eiger1M, tiling){
+                        arr = coordinates_get_square(&detector.shape,
+                                                     &tiling->square);
                 }
                 of(Ufxc, square){
-                        return coordinates_get_square(&detector.shape,
-                                                      square);
+                        arr = coordinates_get_square(&detector.shape,
+                                                     square);
                 }
                 of(Merlin, square){
-                        return coordinates_get_square(&detector.shape,
-                                                      square);
+                        arr = coordinates_get_square(&detector.shape,
+                                                     square);
+                }
+                of(MerlinMedipix3RXQuad, tiling){
+                        arr = coordinates_get_square(&detector.shape,
+                                                     &tiling->square);
                 }
         }
+        return arr;
 }
 
 void hkl_binoculars_detector_2d_coordinates_save(HklBinocularsDetectorEnum n,
@@ -496,30 +505,37 @@ void hkl_binoculars_detector_2d_coordinates_save(HklBinocularsDetectorEnum n,
 uint8_t *hkl_binoculars_detector_2d_mask_get(HklBinocularsDetectorEnum n)
 {
         const struct detector_t detector = get_detector(n);
+        uint8_t *arr = NULL;
 
         match(detector.type){
                 of(ImXpadS70, imxpad){
-                        return mask_get_imxpad(&detector.shape,
-                                               imxpad);
+                        arr = mask_get_imxpad(&detector.shape,
+                                              imxpad);
                 }
                 of(ImXpadS140, imxpad){
-                        return mask_get_imxpad(&detector.shape,
+                        arr = mask_get_imxpad(&detector.shape,
                                                imxpad);
                 }
                 of(XpadFlatCorrected){
-                        return mask_get_xpad_flat_corrected(&detector.shape);
+                        arr = mask_get_xpad_flat_corrected(&detector.shape);
                 }
-                of(Eiger1M, dectris){
-                        return mask_get_dectris(&detector.shape,
-                                                dectris);
+                of(Eiger1M, tiling){
+                        arr = mask_get_tiling(&detector.shape,
+                                              tiling);
                 }
                 of(Ufxc){
-                        return no_mask(&detector.shape);
+                        arr = no_mask(&detector.shape);
                 }
                 of(Merlin){
-                        return no_mask(&detector.shape);
+                        arr = no_mask(&detector.shape);
+                }
+                of(MerlinMedipix3RXQuad, tiling){
+                        arr = mask_get_tiling(&detector.shape,
+                                              tiling);
                 }
         }
+
+        return arr;
 }
 
 uint8_t *hkl_binoculars_detector_2d_mask_load(HklBinocularsDetectorEnum n,
