@@ -13,7 +13,7 @@
  * You should have received a copy of the GNU General Public License
  * along with the hkl library.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Copyright (C) 2003-2021 Synchrotron SOLEIL
+ * Copyright (C) 2003-2022 Synchrotron SOLEIL
  *                         L'Orme des Merisiers Saint-Aubin
  *                         BP 48 91192 GIF-sur-YVETTE CEDEX
  *
@@ -23,7 +23,8 @@
 #include <time.h>
 
 #include "ccan/array_size/array_size.h"
-#include "hkl-binoculars.h"
+#include "datatype99.h"
+#include "hkl-binoculars-private.h"
 #include "hkl-geometry-private.h"
 #include "hkl-matrix-private.h"
 #include "hkl-quaternion-private.h"
@@ -47,13 +48,47 @@ static inline ptrdiff_t max(ptrdiff_t x, ptrdiff_t y)
 	return x ^ ((x ^ y) & -(x < y));
 }
 
-/* Axis */
+/* Axis Limits */
 
-static inline size_t axis_size(const HklBinocularsAxis *self)
+datatype(
+        AxisLimitType,
+        (NoLimit),
+        (Limit, ptrdiff_t)
+        );
+
+struct _HklBinocularsAxisLimits
 {
-	return self->imax - self->imin + 1;
+        AxisLimitType imin;
+        AxisLimitType imax;
+};
+
+void hkl_binoculars_axis_limits_free(HklBinocularsAxisLimits *self)
+{
+	free(self);
 }
 
+HklBinocularsAxisLimits *hkl_binoculars_axis_limits_new(const ptrdiff_t *imin,
+                                                        const ptrdiff_t *imax)
+{
+	HklBinocularsAxisLimits *self = malloc(sizeof(*self));
+
+        if (NULL == imin){
+                self->imin = NoLimit();
+        } else {
+                self->imin = Limit(*imin);
+        }
+
+        if (NULL == imax){
+                self->imax = NoLimit();
+        } else {
+                self->imax = Limit(*imax);
+        }
+
+        return self;
+}
+
+/* Axis */
+
 static inline double axis_min(const HklBinocularsAxis *self)
 {
 	return self->imin * self->resolution;
@@ -193,6 +228,11 @@ static inline void hkl_binoculars_space_item_fprintf(FILE *f, const HklBinocular
                 self->indexes_0[0], self->indexes_0[1], self->indexes_0[2], self->intensity);
 }
 
+static inline int space_is_empty(const HklBinocularsSpace *space)
+{
+        return 0 == darray_size(space->items);
+}
+
 static inline void space_update_axes(HklBinocularsSpace *space,
                                      const char *names[],
                                      size_t n_pixels,
@@ -203,10 +243,13 @@ static inline void space_update_axes(HklBinocularsSpace *space,
         HklBinocularsSpaceItem minimum;
         HklBinocularsSpaceItem maximum;
 
+        if (space_is_empty(space))
+                return;
+
         minimum = maximum = darray_item(space->items, 0);
 
-        darray_foreach(item, space->items){
-                for(i=0; i<darray_size(space->axes); ++i){
+        for(i=0; i<darray_size(space->axes); ++i){
+                darray_foreach(item, space->items){
                         minimum.indexes_0[i] = min(minimum.indexes_0[i], item->indexes_0[i]);
                         maximum.indexes_0[i] = max(maximum.indexes_0[i], item->indexes_0[i]);
                 }
@@ -220,10 +263,113 @@ static inline void space_update_axes(HklBinocularsSpace *space,
         }
 }
 
+static inline int item_in_the_limits(const HklBinocularsSpaceItem *item,
+                                     const HklBinocularsAxisLimits **limits,
+                                     size_t n_limits)
+{
+        int res = TRUE;
+
+        if (NULL != limits){
+                for(size_t i=0; i<n_limits; ++i){
+                        ptrdiff_t v = item->indexes_0[i];
+
+                        match(limits[i]->imin){
+                                of(NoLimit){
+                                }
+                                of(Limit, imin){
+                                        if (v < *imin) res = FALSE;
+                                        break;
+                                }
+                        }
+
+                        match(limits[i]->imax){
+                                of(NoLimit){
+                                }
+                                of(Limit, imax){
+                                        if(v > *imax) res = FALSE;
+                                        break;
+                                }
+                        }
+                }
+        }
+
+        return res;
+}
+
 /* the array is pre filled with the pixel coordinates */
 
-#define HKL_BINOCULARS_SPACE_Q_IMPL(image_t)                            \
-        HKL_BINOCULARS_SPACE_Q_DECL(image_t)                            \
+/* qparqper */
+
+#define HKL_BINOCULARS_SPACE_QPARQPER_IMPL(image_t)                     \
+        HKL_BINOCULARS_SPACE_QPARQPER_DECL(image_t)                     \
+        {                                                               \
+                size_t i;                                               \
+                const char * names[] = {"qpar", "qper"};                \
+                                                                        \
+                assert(ARRAY_SIZE(names) == darray_size(space->axes));  \
+                assert(ARRAY_SIZE(names) == n_resolutions);             \
+                assert(n_pixels == space->max_items);                   \
+                                                                        \
+                const double *q_x = &pixels_coordinates[0 * n_pixels];  \
+                const double *q_y = &pixels_coordinates[1 * n_pixels];  \
+                const double *q_z = &pixels_coordinates[2 * n_pixels];  \
+                                                                        \
+                HklSample *sample = hkl_sample_new("test");             \
+                HklDetector *detector = hkl_detector_factory_new(HKL_DETECTOR_TYPE_0D); \
+                const HklQuaternion q = hkl_geometry_detector_rotation_get(geometry, detector); \
+                const HklVector ki = hkl_geometry_ki_get(geometry);     \
+                double k = hkl_vector_norm2(&ki);                       \
+                HklQuaternion qs_1 = hkl_geometry_sample_rotation_get(geometry, sample); \
+                switch(surf){                                           \
+                case HKL_BINOCULARS_SURFACE_ORIENTATION_VERTICAL:       \
+                {                                                       \
+                        HklQuaternion q_ub;                             \
+                        HklVector v_s = {{1, 0, 0}};                    \
+                        hkl_quaternion_init_from_angle_and_axe(&q_ub, -M_PI_2, &v_s); \
+                        hkl_quaternion_times_quaternion(&qs_1, &q_ub);  \
+                        break;                                          \
+                };                                                      \
+                case HKL_BINOCULARS_SURFACE_ORIENTATION_HORIZONTAL:     \
+                case HKL_BINOCULARS_SURFACE_ORIENTATION_NUM_ORIENTATION: \
+                        break;                                          \
+                };                                                      \
+                hkl_quaternion_conjugate(&qs_1);                        \
+                                                                        \
+                darray_size(space->items) = 0;                          \
+                                                                        \
+                for(i=0;i<n_pixels;++i){                                \
+                        if(NULL == masked || 0 == masked[i]){           \
+                                HklBinocularsSpaceItem item;            \
+                                HklVector v = {{q_x[i], q_y[i], q_z[i]}}; \
+                                                                        \
+                                hkl_vector_rotated_quaternion(&v, &q);  \
+                                hkl_vector_times_double(&v, k);         \
+                                hkl_vector_minus_vector(&v, &ki);       \
+                                hkl_vector_rotated_quaternion(&v, &qs_1); \
+                                                                        \
+                                item.indexes_0[0] = sqrt(v.data[0] * v.data[0] + v.data[1] * v.data[1]) / resolutions[0] / 10; \
+                                item.indexes_0[1] = v.data[2] / resolutions[1] / 10; \
+                                item.intensity = rint((double)image[i] * weight); \
+                                                                        \
+                                if(TRUE == item_in_the_limits(&item, limits, n_limits)) \
+                                        darray_append(space->items, item); \
+                        }                                               \
+                }                                                       \
+                                                                        \
+                space_update_axes(space, names, n_pixels, resolutions); \
+                                                                        \
+                hkl_detector_free(detector);                            \
+                hkl_sample_free(sample);                                \
+        }
+
+HKL_BINOCULARS_SPACE_QPARQPER_IMPL(int32_t);
+HKL_BINOCULARS_SPACE_QPARQPER_IMPL(uint16_t);
+HKL_BINOCULARS_SPACE_QPARQPER_IMPL(uint32_t);
+
+/* qxqyqz */
+
+#define HKL_BINOCULARS_SPACE_QXQYQZ_IMPL(image_t)                            \
+        HKL_BINOCULARS_SPACE_QXQYQZ_DECL(image_t)                            \
         {                                                               \
                 size_t i, j;                                            \
                 const char * names[] = {"qx", "qy", "qz"};              \
@@ -245,12 +391,14 @@ static inline void space_update_axes(HklBinocularsSpace *space,
                 switch(surf){                                           \
                 case HKL_BINOCULARS_SURFACE_ORIENTATION_VERTICAL:       \
                 {                                                       \
-                        HklQuaternion q_ub = {{0, -1, 0, 0}};           \
+                        HklQuaternion q_ub;                             \
+                        HklVector v_s = {{1, 0, 0}};                    \
+                        hkl_quaternion_init_from_angle_and_axe(&q_ub, -M_PI_2, &v_s); \
                         hkl_quaternion_times_quaternion(&qs_1, &q_ub);  \
                         break;                                          \
                 };                                                      \
                 case HKL_BINOCULARS_SURFACE_ORIENTATION_HORIZONTAL:     \
-                case HKL_BINOCULARS_SURFACE_ORIENTATION_NUM_ORIENTATION:\
+                case HKL_BINOCULARS_SURFACE_ORIENTATION_NUM_ORIENTATION: \
                         break;                                          \
                 };                                                      \
                 hkl_quaternion_conjugate(&qs_1);                        \
@@ -273,7 +421,8 @@ static inline void space_update_axes(HklBinocularsSpace *space,
                                 item.intensity = rint((double)image[i] * weight); \
                                                                         \
                                                                         \
-                                darray_append(space->items, item);      \
+                                if(TRUE == item_in_the_limits(&item, limits, n_limits)) \
+                                        darray_append(space->items, item); \
                         }                                               \
                 }                                                       \
                                                                         \
@@ -283,9 +432,11 @@ static inline void space_update_axes(HklBinocularsSpace *space,
                 hkl_sample_free(sample);                                \
         }
 
-HKL_BINOCULARS_SPACE_Q_IMPL(int32_t);
-HKL_BINOCULARS_SPACE_Q_IMPL(uint16_t);
-HKL_BINOCULARS_SPACE_Q_IMPL(uint32_t);
+HKL_BINOCULARS_SPACE_QXQYQZ_IMPL(int32_t);
+HKL_BINOCULARS_SPACE_QXQYQZ_IMPL(uint16_t);
+HKL_BINOCULARS_SPACE_QXQYQZ_IMPL(uint32_t);
+
+/* hkl */
 
 #define HKL_BINOCULARS_SPACE_HKL_IMPL(image_t)                          \
         HKL_BINOCULARS_SPACE_HKL_DECL(image_t)                          \
@@ -325,10 +476,11 @@ HKL_BINOCULARS_SPACE_Q_IMPL(uint32_t);
                                                                         \
                                 for(j=0; j<ARRAY_SIZE(names); ++j){     \
                                         item.indexes_0[j] = rint(v.data[j] / resolutions[j]); \
-                                }\
+                                }                                       \
                                 item.intensity = rint((double)image[i] * weight); \
                                                                         \
-                                darray_append(space->items, item);      \
+                                if(TRUE == item_in_the_limits(&item, limits, n_limits)) \
+                                        darray_append(space->items, item); \
                         }                                               \
                 }                                                       \
                                                                         \
@@ -379,7 +531,7 @@ static inline HklBinocularsCube *empty_cube(const darray_axis *axes)
         return self;
 }
 
-static inline int is_empty(const HklBinocularsCube *self)
+static inline int cube_is_empty(const HklBinocularsCube *self)
 {
         return 0 == darray_size(self->axes);
 }
@@ -429,14 +581,16 @@ void hkl_binoculars_cube_free(HklBinocularsCube *self)
 
 /* Using this method the Cube has already the right dimensions, we
  * just add the Space data into it. */
-void add_space(HklBinocularsCube *cube,
-	       const HklBinocularsSpace *space)
+static inline void add_non_empty_space(HklBinocularsCube *cube,
+                                       const HklBinocularsSpace *space)
 {
 	size_t i;
         size_t n_axes = darray_size(cube->axes);
 	ptrdiff_t lens[n_axes];
         HklBinocularsSpaceItem *item;
 
+        assert(n_axes == darray_size(space->axes));
+
         /* compute the lens */
         lens[0] = 1;
 
@@ -481,26 +635,55 @@ void hkl_binoculars_cube_dims(const HklBinocularsCube *self, size_t ndim, size_t
         }
 }
 
+static inline size_t find_first_non_empty_space_index(size_t n_spaces,
+                                                      const HklBinocularsSpace *const *spaces)
+{
+        size_t i;
+
+        for(i=0; i<n_spaces; ++i)
+                if(!space_is_empty(spaces[i]))
+                        return i;
+
+        return n_spaces;
+}
+
 HklBinocularsCube *hkl_binoculars_cube_new(size_t n_spaces,
                                            const HklBinocularsSpace *const *spaces)
 {
-	size_t i;
-	const HklBinocularsSpace *space0 = spaces[0];
-	HklBinocularsCube *self = empty_cube(&space0->axes);
+        size_t i, i0;
+        HklBinocularsCube *self;
 
-	/* compute the final cube dimensions and the index offset */
-	for(i=1; i<n_spaces; ++i){
-                merge_axes(&self->axes, &spaces[i]->axes);
+        i0 = find_first_non_empty_space_index(n_spaces, spaces);
+
+        if (i0 == n_spaces){ /* all spaces are empty */
+                self = empty_cube(NULL);
+        } else {
+                self = empty_cube(&spaces[i0]->axes);
+
+                /* compute the final cube dimensions and the index offset */
+                for(i=i0; i<n_spaces; ++i){
+                        const HklBinocularsSpace *space = spaces[i];
+
+                        if(space_is_empty(space))
+                                continue;
+
+                        merge_axes(&self->axes, &space->axes);
+                }
+                self->offset0 = compute_offset0(&self->axes);
+
+                /* allocated the final cube photons and contributions */
+                calloc_cube(self);
+
+                /* add all the spaces */
+                for(i=i0; i<n_spaces; ++i){
+                        const HklBinocularsSpace *space = spaces[i];
+
+                        if(space_is_empty(space))
+                                continue;
+
+                        add_non_empty_space(self, space);
+                }
         }
-	self->offset0 = compute_offset0(&self->axes);
-
-	/* allocated the final cube photons and contributions */
-        calloc_cube(self);
-
-	/* add all the spaces */
-	for(i=0; i<n_spaces; ++i){
-		add_space(self, spaces[i]);
-	}
 
 	return self;
 }
@@ -522,14 +705,20 @@ HklBinocularsCube *hkl_binoculars_cube_new_empty_from_cube(const HklBinocularsCu
 
 HklBinocularsCube *hkl_binoculars_cube_new_from_space(const HklBinocularsSpace *space)
 {
-	HklBinocularsCube *self = empty_cube(&space->axes);
+	HklBinocularsCube *self;
 
-	/* allocated the final cube */
-        calloc_cube(self);
+        if(space_is_empty(space)){
+                self = empty_cube(NULL);
+        } else {
+                self = empty_cube(&space->axes);
 
-        add_space(self, space);
+                /* allocated the final cube */
+                calloc_cube(self);
 
-	return self;
+                add_non_empty_space(self, space);
+        }
+
+        return self;
 }
 
 HklBinocularsCube *hkl_binoculars_cube_new_copy(const HklBinocularsCube *src)
@@ -549,13 +738,37 @@ HklBinocularsCube *hkl_binoculars_cube_new_copy(const HklBinocularsCube *src)
         return self;
 }
 
-
-static inline void cube_add_cube(HklBinocularsCube *self,
-                                 const HklBinocularsCube *other)
+static inline void cube_add_cube_2(HklBinocularsCube *self,
+                                   const HklBinocularsCube *other)
 {
-        if ((NULL == other->photons) || (NULL == other->contributions))
-                return;
+        size_t i, j;
+        size_t i_offset, j_offset;
 
+        size_t stride_i = 1;
+        size_t stride_j = stride_i * axis_size(&darray_item(self->axes, 1));
+
+        /* fill the values of other */
+        size_t stride_i_other = 1;
+        size_t stride_j_other = stride_i_other * axis_size(&darray_item(other->axes, 1));
+
+        i_offset = darray_item(other->axes, 1).imin - darray_item(self->axes, 1).imin;
+        j_offset = darray_item(other->axes, 0).imin - darray_item(self->axes, 0).imin;
+
+        for(j=0; j<axis_size(&darray_item(other->axes, 0)); ++j){
+                for(i=0; i<axis_size(&darray_item(other->axes, 1)); ++i){
+                        size_t w = (i + i_offset) * stride_i + (j + j_offset) * stride_j;
+                        size_t w1 = i * stride_i_other + j * stride_j_other;
+
+                        self->photons[w] += other->photons[w1];
+                        self->contributions[w] += other->contributions[w1];
+                }
+        }
+}
+
+
+static inline void cube_add_cube_3(HklBinocularsCube *self,
+                                   const HklBinocularsCube *other)
+{
         size_t i, j, k;
         size_t i_offset, j_offset, k_offset;
 
@@ -585,6 +798,32 @@ static inline void cube_add_cube(HklBinocularsCube *self,
         }
 }
 
+static inline void cube_add_cube(HklBinocularsCube *self,
+                                 const HklBinocularsCube *other)
+{
+        if ((NULL == other->photons) || (NULL == other->contributions))
+                return;
+
+        assert(darray_size(self->axes) == darray_size(other->axes));
+
+        switch(darray_size(self->axes)){
+        case 2: cube_add_cube_2(self, other);
+                break;
+        case 3: cube_add_cube_3(self, other);
+                break;
+        default: assert(0);
+        }
+}
+
+static inline void compute_strides(const darray_axis *axes, size_t strides[], size_t n_strides)
+{
+        size_t i;
+
+        strides[0] = 1;
+        for(i=1; i<n_strides; ++i)
+                strides[i] = strides[i-1] * axis_size(&darray_item(*axes, n_strides - i + 1));
+}
+
 HklBinocularsCube *hkl_binoculars_cube_new_merge(const HklBinocularsCube *cube1,
                                                  const HklBinocularsCube *cube2)
 {
@@ -600,8 +839,8 @@ HklBinocularsCube *hkl_binoculars_cube_new_merge(const HklBinocularsCube *cube1,
         return self;
 }
 
-static void switch_content(HklBinocularsCube *self,
-                           HklBinocularsCube *other)
+static inline void switch_content(HklBinocularsCube *self,
+                                  HklBinocularsCube *other)
 {
         unsigned int *ptr;
         darray_axis tmp;
@@ -634,7 +873,7 @@ void hkl_binoculars_cube_add_space(HklBinocularsCube *self,
         hkl_binoculars_cube_fprintf(stdout, self);
         hkl_binoculars_space_fprintf(stdout, space);
 #endif
-        if(is_empty(self)){
+        if(cube_is_empty(self)){
 #ifdef DEBUG
                 fprintf(stdout, "\nCreate a new empty cube");
 #endif
@@ -644,22 +883,24 @@ void hkl_binoculars_cube_add_space(HklBinocularsCube *self,
                 hkl_binoculars_cube_free(cube);
         }else{
                 /* check the compatibility of the cube and the space. */
-                if (does_not_include(&self->axes, &space->axes)){
+                if (!space_is_empty(space)){
+                        if (does_not_include(&self->axes, &space->axes)){
 #ifdef DEBUG
-                        fprintf(stdout, "\nthe Cube does not contain the space, so create a new cube.");
+                                fprintf(stdout, "\nthe Cube does not contain the space, so create a new cube.");
 #endif
-                        HklBinocularsCube *cube = empty_cube(&self->axes);
+                                HklBinocularsCube *cube = empty_cube(&self->axes);
 
-                        merge_axes(&cube->axes, &space->axes); /* circonscript */
-                        cube->offset0 = compute_offset0(&cube->axes);
-                        calloc_cube(cube);
+                                merge_axes(&cube->axes, &space->axes); /* circonscript */
+                                cube->offset0 = compute_offset0(&cube->axes);
+                                calloc_cube(cube);
 
-                        cube_add_cube(cube, self);
+                                cube_add_cube(cube, self);
 
-                        switch_content(self, cube);
-                        hkl_binoculars_cube_free(cube);
+                                switch_content(self, cube);
+                                hkl_binoculars_cube_free(cube);
+                        }
+                        add_non_empty_space(self, space);
                 }
-                add_space(self, space);
         }
 #ifdef DEBUG
         fprintf(stdout, "\n");
