@@ -13,7 +13,7 @@
  * You should have received a copy of the GNU General Public License
  * along with the hkl library.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Copyright (C) 2003-2021 Synchrotron SOLEIL
+ * Copyright (C) 2003-2022 Synchrotron SOLEIL
  *                         L'Orme des Merisiers Saint-Aubin
  *                         BP 48 91192 GIF-sur-YVETTE CEDEX
  *
@@ -39,6 +39,7 @@ typedef enum _HklBinocularsDetectorEnum
         HKL_BINOCULARS_DETECTOR_DECTRIS_EIGER1M,
         HKL_BINOCULARS_DETECTOR_UFXC,
         HKL_BINOCULARS_DETECTOR_MERLIN,
+        HKL_BINOCULARS_DETECTOR_MERLIN_MEDIPIX_3RX_QUAD,
         /* Add new your detectors here */
         HKL_BINOCULARS_DETECTOR_NUM_DETECTORS,
 } HklBinocularsDetectorEnum;
@@ -68,21 +69,23 @@ HKLAPI extern void hkl_binoculars_detector_2d_sixs_calibration(HklBinocularsDete
                                                                int width, int height,
                                                                int ix0, int iy0, double sdd,
                                                                double detrot);
+
+/***********/
+/* Limits  */
+/***********/
+
+typedef struct _HklBinocularsAxisLimits HklBinocularsAxisLimits;
+
+HKLAPI extern void hkl_binoculars_axis_limits_free(HklBinocularsAxisLimits *self);
+
+HKLAPI extern HklBinocularsAxisLimits *hkl_binoculars_axis_limits_new(const ptrdiff_t *min,
+                                                                      const ptrdiff_t *max);
+
 /********/
 /* Axis */
 /********/
 
 typedef struct _HklBinocularsAxis HklBinocularsAxis;
-struct _HklBinocularsAxis
-{
-	const char *name; /* the name of the axis */
-	size_t index; /* the index of the axis of the projection used */
-	double resolution; /* the resolution of the bins */
-	ptrdiff_t imin; /* the minimum index of the axis min = imin * resolution */
-	ptrdiff_t imax; /* the maximum index of the axis max = imax * resolution */
-};
-
-typedef darray(HklBinocularsAxis) darray_axis;
 
 HKLAPI extern double *hkl_binoculars_axis_array(const HklBinocularsAxis *self);
 
@@ -91,22 +94,7 @@ HKLAPI extern double *hkl_binoculars_axis_array(const HklBinocularsAxis *self);
 /* Space */
 /*********/
 
-typedef struct _HklBinocularsSpaceItem HklBinocularsSpaceItem;
-struct _HklBinocularsSpaceItem
-{
-        ptrdiff_t indexes_0[3]; /* for now hardcode the max number of axes */
-        uint32_t intensity;
-};
-
-typedef darray(HklBinocularsSpaceItem) darray_HklBinocularsSpaceItem;
-
 typedef struct _HklBinocularsSpace HklBinocularsSpace;
-struct _HklBinocularsSpace
-{
-        darray_axis axes;
-        size_t max_items;
-        darray_HklBinocularsSpaceItem items;
-};
 
 typedef enum _HklBinocularsSurfaceOrientationEnum
 {
@@ -122,55 +110,11 @@ HKLAPI extern HklBinocularsSpace *hkl_binoculars_space_new(size_t n_indexes_0,
 
 HKLAPI extern void hkl_binoculars_space_free(HklBinocularsSpace *self);
 
-#define HKL_BINOCULARS_SPACE_Q_DECL(image_t)                            \
-        void hkl_binoculars_space_q_ ## image_t (HklBinocularsSpace *space, \
-                                                 const HklGeometry *geometry, \
-                                                 const image_t *image,  \
-                                                 size_t n_pixels,       \
-                                                 double weight,         \
-                                                 const double *pixels_coordinates, \
-                                                 size_t pixels_coordinates_ndim, \
-                                                 const size_t *pixels_coordinates_dims, \
-                                                 const double *resolutions, \
-                                                 size_t n_resolutions,  \
-                                                 const uint8_t *masked, \
-                                                 HklBinocularsSurfaceOrientationEnum surf)
-
-HKLAPI extern HKL_BINOCULARS_SPACE_Q_DECL(int32_t);
-HKLAPI extern HKL_BINOCULARS_SPACE_Q_DECL(uint16_t);
-HKLAPI extern HKL_BINOCULARS_SPACE_Q_DECL(uint32_t);
-
-
-#define HKL_BINOCULARS_SPACE_HKL_DECL(image_t)\
-        void hkl_binoculars_space_hkl_ ## image_t (HklBinocularsSpace *space,\
-                                                   const HklGeometry *geometry, \
-                                                   const HklSample *sample, \
-                                                   const image_t *image, \
-                                                   size_t n_pixels,     \
-                                                   double weight,       \
-                                                   const double *pixels_coordinates, \
-                                                   size_t pixels_coordinates_ndim, \
-                                                   const size_t *pixels_coordinates_dims, \
-                                                   const double *resolutions, \
-                                                   size_t n_resolutions, \
-                                                   const uint8_t *masked)
-
-HKLAPI extern HKL_BINOCULARS_SPACE_HKL_DECL(int32_t);
-HKLAPI extern HKL_BINOCULARS_SPACE_HKL_DECL(uint16_t);
-HKLAPI extern HKL_BINOCULARS_SPACE_HKL_DECL(uint32_t);
-
 /********/
 /* Cube */
 /********/
 
 typedef  struct _HklBinocularsCube HklBinocularsCube;
-struct _HklBinocularsCube
-{
-        darray_axis axes;
-        ptrdiff_t offset0;
-	unsigned int *photons;
-	unsigned int *contributions;
-};
 
 HKLAPI extern void hkl_binoculars_cube_free(HklBinocularsCube *self);
 
@@ -192,6 +136,79 @@ HKLAPI extern HklBinocularsCube *hkl_binoculars_cube_new_merge(const HklBinocula
 HKLAPI extern void hkl_binoculars_cube_add_space(HklBinocularsCube *self,
                                                  const HklBinocularsSpace *space);
 
+HKLAPI extern void hkl_binoculars_cube_save_hdf5(const char *fn,
+                                                 const HklBinocularsCube *self);
+
+/***************/
+/* Projections */
+/***************/
+
+/* qperqpar */
+
+#define HKL_BINOCULARS_SPACE_QPARQPER_DECL(image_t)                            \
+        void hkl_binoculars_space_qparqper_ ## image_t (HklBinocularsSpace *space, \
+                                                        const HklGeometry *geometry, \
+                                                        const image_t *image, \
+                                                        size_t n_pixels, \
+                                                        double weight,  \
+                                                        const double *pixels_coordinates, \
+                                                        size_t pixels_coordinates_ndim, \
+                                                        const size_t *pixels_coordinates_dims, \
+                                                        const double *resolutions, \
+                                                        size_t n_resolutions, \
+                                                        const uint8_t *masked, \
+                                                        HklBinocularsSurfaceOrientationEnum surf, \
+                                                        const HklBinocularsAxisLimits **limits, \
+                                                        size_t n_limits)
+
+HKLAPI extern HKL_BINOCULARS_SPACE_QPARQPER_DECL(int32_t);
+HKLAPI extern HKL_BINOCULARS_SPACE_QPARQPER_DECL(uint16_t);
+HKLAPI extern HKL_BINOCULARS_SPACE_QPARQPER_DECL(uint32_t);
+
+/* qxqyqz */
+
+#define HKL_BINOCULARS_SPACE_QXQYQZ_DECL(image_t)                            \
+        void hkl_binoculars_space_qxqyqz_ ## image_t (HklBinocularsSpace *space, \
+                                                      const HklGeometry *geometry, \
+                                                      const image_t *image, \
+                                                      size_t n_pixels,  \
+                                                      double weight,    \
+                                                      const double *pixels_coordinates, \
+                                                      size_t pixels_coordinates_ndim, \
+                                                      const size_t *pixels_coordinates_dims, \
+                                                      const double *resolutions, \
+                                                      size_t n_resolutions, \
+                                                      const uint8_t *masked, \
+                                                      HklBinocularsSurfaceOrientationEnum surf, \
+                                                      const HklBinocularsAxisLimits **limits, \
+                                                      size_t n_limits)
+
+HKLAPI extern HKL_BINOCULARS_SPACE_QXQYQZ_DECL(int32_t);
+HKLAPI extern HKL_BINOCULARS_SPACE_QXQYQZ_DECL(uint16_t);
+HKLAPI extern HKL_BINOCULARS_SPACE_QXQYQZ_DECL(uint32_t);
+
+/* hkl */
+
+#define HKL_BINOCULARS_SPACE_HKL_DECL(image_t)                          \
+        void hkl_binoculars_space_hkl_ ## image_t (HklBinocularsSpace *space, \
+                                                   const HklGeometry *geometry, \
+                                                   const HklSample *sample, \
+                                                   const image_t *image, \
+                                                   size_t n_pixels,     \
+                                                   double weight,       \
+                                                   const double *pixels_coordinates, \
+                                                   size_t pixels_coordinates_ndim, \
+                                                   const size_t *pixels_coordinates_dims, \
+                                                   const double *resolutions, \
+                                                   size_t n_resolutions, \
+                                                   const uint8_t *masked, \
+                                                   const HklBinocularsAxisLimits **limits, \
+                                                   size_t n_limits)
+
+HKLAPI extern HKL_BINOCULARS_SPACE_HKL_DECL(int32_t);
+HKLAPI extern HKL_BINOCULARS_SPACE_HKL_DECL(uint16_t);
+HKLAPI extern HKL_BINOCULARS_SPACE_HKL_DECL(uint32_t);
+
 G_END_DECLS
 
 #endif
